package org.processmining.plugins.bpsim;

import java.rmi.server.UID;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.NodeID;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.plugins.bpmn.BpmnDefinitions;
import org.processmining.plugins.bpmn.BpmnDefinitions.BpmnDefinitionsBuilder;
import org.processmining.plugins.bpmn.BpmnProcess;
import org.processmining.plugins.bpsim.constantparameter.BooleanParameter;
import org.processmining.plugins.bpsim.constantparameter.DateTimeParameter;
import org.processmining.plugins.bpsim.constantparameter.DurationParameter;
import org.processmining.plugins.bpsim.constantparameter.FloatingParameter;
import org.processmining.plugins.bpsim.constantparameter.NumericParameter;
import org.processmining.plugins.bpsim.constantparameter.StringParameter;
import org.processmining.plugins.bpsim.distributionparameter.BetaDistribution;
import org.processmining.plugins.bpsim.distributionparameter.ErlangDistribution;
import org.processmining.plugins.bpsim.distributionparameter.GammaDistribution;
import org.processmining.plugins.bpsim.distributionparameter.LogNormalDistribution;
import org.processmining.plugins.bpsim.distributionparameter.NegativeExponentialDistribution;
import org.processmining.plugins.bpsim.distributionparameter.NormalDistribution;
import org.processmining.plugins.bpsim.distributionparameter.TruncatedNormalDistribution;
import org.processmining.plugins.bpsim.distributionparameter.UniformDistribution;
import org.processmining.plugins.bpsim.distributionparameter.UserDistribution;
import org.processmining.plugins.bpsim.distributionparameter.UserDistributionDataPoint;
import org.processmining.plugins.bpsim.distributionparameter.WeibullDistribution;
import org.processmining.plugins.bpsim.elementparameter.ControlParameters;
import org.processmining.plugins.bpsim.elementparameter.CostParameters;
import org.processmining.plugins.bpsim.elementparameter.ExtendedParameters;
import org.processmining.plugins.bpsim.elementparameter.ResourceParameters;
import org.processmining.plugins.bpsim.elementparameter.TimeParameters;
import org.processmining.plugins.bpsim.enumerations.SimulationTool;
import org.processmining.plugins.bpsim.extended.DataValueMatrixValue;
import org.processmining.plugins.bpsim.parameter.BPSimCalendar;
import org.processmining.plugins.bpsim.parameter.ExpressionParameter;
import org.processmining.plugins.bpsim.parameter.Parameter;
import org.processmining.plugins.bpsim.parameter.ParameterValue;
import org.processmining.plugins.bpsim.parameter.ResultType;
import org.processmining.plugins.bpsim.scenario.BPSim;
import org.processmining.plugins.bpsim.scenario.BPSimData;
import org.processmining.plugins.bpsim.scenario.ElementParameters;
import org.processmining.plugins.bpsim.scenario.PropertyParameters;
import org.processmining.plugins.bpsim.scenario.Scenario;
import org.processmining.plugins.bpsim.scenario.ScenarioParameters;
import org.processmining.plugins.bpsim.scenario.TimeUnit;
import org.processmining.plugins.bpsim.scenario.VendorExtension;
import org.processmining.plugins.converters.ProcessTree2BPMNConverter;
import org.processmining.plugins.properties.enumerations.OROptions;
import org.processmining.plugins.properties.enumerations.TimeOptions;
import org.processmining.plugins.properties.processmodel.Property;
import org.processmining.plugins.properties.processmodel.TimeValue;
import org.processmining.plugins.properties.processmodel.abstractproperty.TimeValueImpl;
import org.processmining.plugins.properties.processmodel.impl.Distribution;
import org.processmining.plugins.properties.processmodel.impl.NrOfResources;
import org.processmining.plugins.properties.processmodel.impl.SimulationProperty;
import org.processmining.plugins.properties.processmodel.impl.TaskDistribution;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.Originator;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTreeElement;
import org.processmining.processtree.Task;
import org.processmining.processtree.Variable;

import biweekly.ICalendar;
import biweekly.component.VEvent;
import biweekly.property.DateEnd;
import biweekly.property.DateStart;
import biweekly.util.Duration;
import biweekly.util.Recurrence;
import biweekly.util.Recurrence.Frequency;


@Plugin(
		name = "Process Tree to BPSim converter", 
		parameterLabels = {"Process Tree"}, 
		returnLabels = { "BPSim"}, 
		returnTypes = { BPSim.class}, 
		userAccessible = true, 
		help = "Converts a Process Tree into a BPSim model"
	)

public class ProcessTree2BPSimConverter {		
	
	private BPMNDiagram globalbpmnDiagram;
	private BpmnDefinitions globalBpmnDefinitions;
	private Map<NodeID, UUID> referenceMap;
	private String bpmnTreeId;
	private UUID treeId;
	private SimulationTool simTool;
	
	@UITopiaVariant(
			affiliation = UITopiaVariant.EHV, 
			author = "T.C. van der Hoek", 
			email = "t.c.v.d.hoek@student.tue.nl"
		)
	@PluginVariant(variantLabel = "Convert Porcess Tree to BPSim", requiredParameterLabels = { 0 })
	public BPSim converterDefaultPlugin(PluginContext context, ProcessTree tree) throws IllegalAccessException, InstantiationException {			    			
		return converterPlugin(context, tree);
	}
	
	public Map<NodeID, UUID> getMapping(){
		return referenceMap;
	}
				
	private BPSim converterPlugin(PluginContext context, ProcessTree tree) throws IllegalAccessException, InstantiationException {		
		Progress progress = context.getProgress();
		progress.setCaption("Converting Process Tree to BPSim");
		
		//First the process tree is converted to a BPMN diagram
		ProcessTree2BPMNConverter pt2bpmn = new ProcessTree2BPMNConverter();
		Object[] bpmnConversionResult = pt2bpmn.convert(context, tree, false);
		BPMNDiagram bpmnDiagram = (BPMNDiagram)bpmnConversionResult[0];
		@SuppressWarnings("unchecked")
		Map<NodeID, UUID> refMap = (Map<NodeID, UUID>)bpmnConversionResult[1];
		
		globalbpmnDiagram = bpmnDiagram;
		referenceMap = refMap;
		// TODO make a proper argument out of this
		simTool = SimulationTool.LSim;
		
		//Next the properties are converted to BPSim elements
		BPSim bpsim = new BPSim();
		
		bpsim.setBpmnDiagram(bpmnDiagram);
		
		BpmnDefinitionsBuilder definitionsBuilder = new BpmnDefinitionsBuilder(context, bpmnDiagram);
		BpmnDefinitions definitions = new BpmnDefinitions("definitions", definitionsBuilder);
		bpsim.setBpmnDefinitions(definitions);
		globalBpmnDefinitions = definitions;
		for(BpmnProcess proc: definitions.getProcesses()){
			// we should only have 1, so lets just set the id
			bpmnTreeId = proc.getId();
			treeId = tree.getID();
		}
		
		BPSimData bpsimData = new BPSimData();									
		convert(tree, bpsimData);		
		bpsim.setBpsimData(bpsimData);
				
		return bpsim;
	}	
	
	private void convert(ProcessTree tree, BPSimData bpsimData) throws IllegalAccessException, InstantiationException {		
		
		//create bpsim scenario
		Scenario sc = new Scenario();				
		sc.setName(tree.getName());
		Date start = new Date();
		start.setHours(0);
		start.setMinutes(0);
		start.setSeconds(0);
		start.setDate(1);
		start.setMonth(1);
		start.setYear(115);
		sc.setCreated(start);
		sc.setModified(start);
	
		treeWalkthrough(tree, sc);
		
		Collection<Scenario> scs = bpsimData.getScenarios();
		scs.add(sc);		
		bpsimData.setScenarios(scs);
				
	}
		
	private void treeWalkthrough(ProcessTree tree, Scenario sc) throws IllegalAccessException, InstantiationException {
		//go through all the tree elements
		if (!tree.getIndependentProperties().keySet().isEmpty()){
			convertProperty(tree, sc);
		}								
		
		for (Node node: tree.getNodes()) {
			if (!node.getIndependentProperties().keySet().isEmpty()){
				convertProperty(node, sc);
			}
			// Probabilities
			// we need to convert the probability for the expression
			Auxiliary.setProbabilityElement(globalbpmnDiagram, node, sc, this);
		}
		
		for (Edge edge: tree.getEdges()) {
			if (!edge.getIndependentProperties().keySet().isEmpty()){
				convertProperty(edge, sc);
			}
		}
		
		for (Originator originator: tree.getOriginators()) {
			if (!originator.getIndependentProperties().keySet().isEmpty()){
				convertProperty(originator, sc);
			}
		}
		
		for (Variable var: tree.getVariables()) {
			if (!var.getIndependentProperties().keySet().isEmpty()){
				convertProperty(var, sc);
			}
		}
		
		for (Expression expr: tree.getExpressions()) {
			if (!expr.getIndependentProperties().keySet().isEmpty()){
				convertProperty(expr, sc);
			}
		}		
	}

	//***********************************************************************************
	private void convertProperty(ProcessTreeElement treeNode, Scenario sc) throws IllegalAccessException, InstantiationException {
		if (!treeNode.getIndependentProperties().keySet().isEmpty()){					
			//create new elementparameter			
			ElementParameters ep = new ElementParameters();
			
			//set references
			Object[] refLink = getBPMNElement(treeNode.getID());
			boolean linkFound = (Boolean)refLink[1];
			if(simTool.equals(SimulationTool.LSim) && treeNode instanceof Originator){
				linkFound = false;
				ep.setElementRefnodeID(treeNode.getID().toString());
			}
			if (linkFound) {
				String linkedID = (String)refLink[0];
				ep.setElementRefnodeID(linkedID);
			}
			if(!(treeNode instanceof ProcessTree)){
				ep.setElementRefID(treeNode.getID());
			}
			boolean bpsimParameter = false;
			// also add some counters to this element, and to the tree
			if(treeNode instanceof Node){
				ControlParameters cParam;
				if(ep.getControlParameters() != null){
					cParam = ep.getControlParameters();
				}
				else{
					cParam = new ControlParameters();
				}
				Parameter p;
				if(cParam.getTriggerCount() != null){
					p = cParam.getTriggerCount();
				}
				else{
					p = new Parameter();
				}
				List<ResultType> lresType = new ArrayList<ResultType>();
				lresType.add(ResultType.count);
				p.setResultRequest(lresType);
				cParam.setTriggerCount(p);
				ep.setControlParameters(cParam);
				bpsimParameter = true;
			}
			if(treeNode instanceof ProcessTree){
				// add it manually since the arrival process overwrites the id
				ElementParameters epTree = new ElementParameters();
				epTree.setElementRefID(treeNode.getID());
				epTree.setElementRefnodeID(bpmnTreeId);
				ControlParameters cParam;
				if(epTree.getControlParameters() != null){
					cParam = epTree.getControlParameters();
				}
				else{
					cParam = new ControlParameters();
				}
				Parameter p;
				if(cParam.getTriggerCount() != null){
					p = cParam.getTriggerCount();
				}
				else{
					p = new Parameter();
				}
				List<ResultType> lresType = new ArrayList<ResultType>();
				lresType.add(ResultType.count);
				p.setResultRequest(lresType);
				cParam.setTriggerCount(p);
				epTree.setControlParameters(cParam);
				// also get some timing information
				TimeParameters tParam;
				if(epTree.getTimeParameters() != null){
					tParam = epTree.getTimeParameters();
				}
				else{
					tParam = new TimeParameters();
				}
				if(tParam.getProcessingTime() != null){
					p = tParam.getProcessingTime();
				}
				else{
					p = new Parameter();
				}
				lresType = new ArrayList<ResultType>();
				lresType.add(ResultType.mean);
				p.setResultRequest(lresType);
				tParam.setProcessingTime(p);
				epTree.setTimeParameters(tParam);
				
				// now also ask for the vendor extension of the throughput time
				VendorExtension venExtension = new VendorExtension();
				venExtension.setName("l-sim-process-results-v1");
				venExtension.setValue("<AverageTime><ResultRequest>mean</ResultRequest></AverageTime>");
				Collection<VendorExtension> venExtensions;
				if(epTree.getVendorExtension() != null){
					venExtensions = ep.getVendorExtension();
				}
				else{
					venExtensions = new ArrayList<>();
				}
				venExtensions.add(venExtension);
				epTree.setVendorExtension(venExtensions);
				
				Collection<ElementParameters> eps = sc.getElementParameter();				
				eps.add(epTree);
				sc.setElementParameter(eps);
				
			}
			if(simTool.equals(SimulationTool.LSim) && treeNode instanceof Task.Manual){
				// we need to add the selection criteria
				ResourceParameters resPar = ep.getResourceParameters();
				if(resPar == null){
					resPar = new ResourceParameters();
				}
				Parameter select = new Parameter();
				ExpressionParameter exPar = new ExpressionParameter();
				String expr = "bpsim:";
				// now we have to iterate through the resources
				boolean first = true;
				// first do a check if there are multiple resources
				
				if(((Task.Manual)treeNode).getOriginators().size() > 1){
					expr += "orResource(";
				}
				for(Originator org: ((Task.Manual) treeNode).getOriginators()){
					if(!first){
						expr += ", ";
					}
					expr += "getResource(&quot;" + org.getID() + "&quot;, 1)";
					
					first = false;
				}
				if(((Task.Manual)treeNode).getOriginators().size() > 1){
					expr += ")";
				}
				exPar.setValue(expr);
				Collection<ParameterValue> pvs = select.getValue();
				pvs.add(exPar);
				select.setValue(pvs);
				resPar.setSelection(select);
				ep.setResourceParameters(resPar);				
				bpsimParameter = true;
			}			
			for ( Property<?> prop: treeNode.getIndependentProperties().keySet()){				
				//System.out.println(prop.getName());
				switch(prop.getID().intValue()) {
					case (int)-5090854327664326825L : { //arrival process
						convertArrivalProcess(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-5148631613566217221L : { //automatic distribution
						convertAutomaticDistr(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-8455197641724609928L : { //costs
						convertCosts(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)8070362200740687215L : { //event distribution
						convertEventDistr(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)6697705505907948666L : { //event time
						convertEventTime(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)6858242719283578195L: { //initial data value
						if(simTool.equals(SimulationTool.LSim)){
							// not supported
						}
						else{
							convertinitialDataValue(prop, treeNode, sc);	
							bpsimParameter = true;
						}
						break;
					}
					case (int)845428618399631425L : { //nr of resources
						if(!simTool.equals(SimulationTool.LSim)){
							convertNrOfResources(prop, treeNode, ep);							
							bpsimParameter = true;
						}
						break;
					}
					case (int)-7544999668426095062L : { //processing time
						convertProcessingTime(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-6741244140667692872L : { //processing time order
						convertProcessingTimeOrder(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-1251672671121390315L : { //queue time
						convertQueueTime(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-8846407474328210618L : { //queue time order
						convertQueueTimeOrder(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)3142705774461179331L : { //simulation property
						convertSimulationProperty(prop, treeNode, sc);	
						//bpsimParameter = true;
						break;
					}					
					case (int)-4885873430234716452L : { //work schedule
						convertWorkSchedule(prop, treeNode, ep, sc);
						bpsimParameter = true;
						break;
					}
					case (int)-8348569829190729351L: { //yerkes dodson interval
						convertYerkesDodsonInterval(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)3852971573640946989L:
						convertTaskDistribution(prop, treeNode, ep);
						bpsimParameter = true;
						break;
				}				
				switch(prop.getID().intValue()) {
					//EXTENDED BPSim property conversion----------------------------------------------------------------
					case (int)-118771733070284461L: { //data value matrix
						convertDataValueMatrix(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-6489711900627424587L: { //data values
						convertDataValues(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-3304239541357626018L: { //fitness
						convertFitness(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}					
					case (int)748599691379745184L: { //or semantic
						convertORSemantics(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)4893070190920165692L: { //throughput time
						convertThroughputTime(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-8174008339248178700L: { //throughput time order
						convertThroughputTimeOrder(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)-5970506923461775924L: { //utilisation
						convertUtilisation(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)1207196437601255661L: { //wait time
						convertWaitTime(prop, treeNode, ep);
						bpsimParameter = true;
						break;
					}
					case (int)1200976078296640748L: { //wait time order
						convertWaitTimeOrder(prop, treeNode, ep);	
						bpsimParameter = true;
						break;
					}														
				}
			}
			
			//add elementparameter only if used
			if (bpsimParameter) {
				Collection<ElementParameters> eps = sc.getElementParameter();				
				eps.add(ep);
				sc.setElementParameter(eps);
			}
		}
	}
	
	private void convertArrivalProcess(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//set the reference to the start bpmn element		
		for(Event event: globalbpmnDiagram.getEvents()) {
			if (event.getLabel() == "Start") {
				ep.setElementRefnodeID(event.getId().toString());
			}
		}
		
		//get or create parameters
		ControlParameters cps;		
		if (ep.getControlParameters() == null) {
			cps = new ControlParameters();			
		} else {
			cps = ep.getControlParameters();			
		}		
		
		Parameter p;
		if (cps.getInterTriggerTimer() == null) {
			p = new Parameter();
		} else {
			p = cps.getInterTriggerTimer();
		}
			
		//set intertrigger time to p		
		Collection<ParameterValue> pvs = p.getValue();		
		
		ParameterValue pv = convertDistribution((Distribution)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);
				
		cps.setInterTriggerTimer(p);
		ep.setControlParameters(cps);		
	}
	
	private void convertAutomaticDistr(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();
		} else {
			tps = ep.getTimeParameters();		
		}	
		
		Parameter p;
		if (tps.getProcessingTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getProcessingTime();
		}
			
		//set processing time to p		
		Collection<ParameterValue> pvs = p.getValue();		
					
		ParameterValue pv = convertDistribution((Distribution)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);
		
		List<ResultType> lresType = new ArrayList<ResultType>();
		lresType.add(ResultType.mean);
		p.setResultRequest(lresType);
		
		tps.setProcessingTime(p);
		ep.setTimeParameters(tps);				
	}
	
	private void convertCosts(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws NumberFormatException, IllegalAccessException, InstantiationException {
		//get or create parameters
		CostParameters cps;		
		if (ep.getCostParameters() == null) {
			cps = new CostParameters();			
		} else {
			cps = ep.getCostParameters();			
		}
		
		Parameter p;
		if (cps.getUnitCost() == null) {
			p = new Parameter();
		} else {
			p = cps.getUnitCost();
		}
		
		//set costs to p
		Collection<ParameterValue> pvs = p.getValue();
		
		FloatingParameter pv = new FloatingParameter();
				
		pv.setValue(Float.valueOf(treeNode.getIndependentProperty(prop).toString()));	
		pvs.add(pv);
		p.setValue(pvs);
		
		cps.setUnitCost(p);
		ep.setCostParameters(cps);
	}
		
	private void convertEventDistr(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		ControlParameters cp;
		if (ep.getControlParameters() == null){
			cp = new ControlParameters();
		}
		else{
			cp = ep.getControlParameters();
		}
		
		Parameter p;
		if(cp.getInterTriggerTimer() == null){
			p = new Parameter();
		}
		else{
			p = cp.getInterTriggerTimer();
		}
		
		// set the distribution
		Collection<ParameterValue> pvs = p.getValue();
		ParameterValue pv = convertDistribution((Distribution)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);
		List<ResultType> resReq = new ArrayList<>();
		resReq.add(ResultType.count);
		p.setResultRequest(resReq);
		
		cp.setInterTriggerTimer(p);
		ep.setControlParameters(cp);
		
		/*
		TimeParameters tps;
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
		
		Parameter p;
		if (tps.getTransferTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getTransferTime();
		}
		
		//set distribution time to p		
		Collection<ParameterValue> pvs = p.getValue();		
					
		ParameterValue pv = convertDistribution((Distribution)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);
		tps.setTransferTime(p);
		ep.setTimeParameters(tps);
		*/	
	}
	
	private void convertTaskDistribution(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
		
		Parameter p;
		if (tps.getTransferTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getTransferTime();
		}
		
		//set distribution time to p		
		Collection<ParameterValue> pvs = p.getValue();		
					
		ParameterValue pv = convertDistribution((Distribution)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);
		List<ResultType> resReq = new ArrayList<>();
		resReq.add(ResultType.mean);
		p.setResultRequest(resReq);
		
		tps.setProcessingTime(p);
		ep.setTimeParameters(tps);	
	}
	
	private void convertEventTime(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
		
		Parameter p;
		if (tps.getTransferTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getTransferTime();
		}
		
		//set event time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);
						
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
		
		p.setValue(pvs);
		List<ResultType> lresType = new ArrayList<ResultType>();
		lresType.add(ResultType.mean);
		p.setResultRequest(lresType);
		tps.setProcessingTime(p);
		
		if(simTool.equals(SimulationTool.LSim)){
			
		}
		else{
			// now also collect some extra information about the queue time
			if (tps.getQueueTime() == null) {
				p = new Parameter();
			} else {
				p = tps.getQueueTime();
			}
			lresType = new ArrayList<ResultType>();
			lresType.add(ResultType.mean);
			p.setResultRequest(lresType);
			tps.setQueueTime(p);
		}
		
		ep.setTimeParameters(tps);		
	}
	
	private void convertinitialDataValue(Property<?> prop, ProcessTreeElement treeNode, Scenario sc) throws IllegalAccessException, InstantiationException {					
		//set the reference to the start bpmn element		
		NodeID startEventId = null;
		for(Event event: globalbpmnDiagram.getEvents()) {
			if (event.getLabel() == "Start") {
				startEventId = event.getId();
				break;
			}
		}
		boolean old = false;
		ElementParameters ep = null;
		if (startEventId != null) {			
			for(ElementParameters epTemp: sc.getElementParameter()) {				
				if (epTemp.getElementRefnodeID().equalsIgnoreCase(startEventId.toString())){					
					ep = epTemp;
					old = true;
					break;
				}
			}
			if (ep == null) {				
				ep = new ElementParameters();
				// TODO: start should not be linked to the tree
				//ep.setElementRefID(treeNode.getID());
				ep.setElementRefnodeID(startEventId.toString());
			}
		} else {			
			ep = new ElementParameters();
			// TODO: should not be linked to the tree
			//ep.setElementRefID(treeNode.getID());			
		}
		
		//get or create parameters
		PropertyParameters eps;		
		if (ep.getPropertyParameters() == null) {
			eps = new PropertyParameters();			
		} else {
			eps = ep.getPropertyParameters();			
		}			
	
		Map<String, Parameter> map;
		if (eps.getProperty() == null) {
			map = new HashMap<String, Parameter>();
		} else {
			map = eps.getProperty();
		}
		
		Parameter p = new Parameter();
		Collection<ParameterValue> pvs = p.getValue();
		
		StringParameter pv = new StringParameter();
		pv.setValue((String)treeNode.getIndependentProperty(prop));
		pvs.add(pv);
		p.setValue(pvs);				
		
		map.put(treeNode.getName(), p);
		
		eps.setProperty(map);

		ep.setPropertyParameters(eps);
		
		if (!old) {			
			Collection<ElementParameters> elemPara = sc.getElementParameter();
			elemPara.add(ep);
			sc.setElementParameter(elemPara);
		}
		
	}
	
	private void convertNrOfResources(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters		
		ResourceParameters rps;		
		if (ep.getResourceParameters() == null) {
			rps = new ResourceParameters();			
		} else {
			rps = ep.getResourceParameters();			
		}	
		
		Parameter p;
		if (rps.getQuantity() == null) {
			p = new Parameter();
		} else {
			p = rps.getQuantity();
		}
	
		Collection<ParameterValue> pvs = p.getValue();
		
		NumericParameter pv = new NumericParameter();
		
		pv.setValue(Long.valueOf(treeNode.getIndependentProperty(prop).toString()));	
		pvs.add(pv);
		p.setValue(pvs);
		
		rps.setQuantity(p);
		ep.setResourceParameters(rps);
	}
	
	private void convertProcessingTime(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		TimeParameters tps;	
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}	
		
		Parameter p;
		if (tps.getProcessingTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getProcessingTime();
		}
			
		//set processing time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);
		
		UserDistribution pv = new UserDistribution();							
		pv.setDiscrete(true);
				
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
				
		p.setValue(pvs);
			
		tps.setProcessingTime(p);
				
		ep.setTimeParameters(tps);	
	}
	
	private void convertProcessingTimeOrder(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
		
		Parameter p;
		if (tps.getProcessingTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getProcessingTime();
		}
			
		//set processing time to p		
		Collection<ParameterValue> pvs = p.getValue();		
		
		@SuppressWarnings("unchecked")
		ArrayList<Double> values = (ArrayList<Double>) treeNode.getIndependentProperty(prop);
		float probability = 1f / values.size();
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (int i=0;i<values.size();i++){							
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
						
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);
			floatValue.setInstance(new UID().toString() + "-" + (i));
			floatValue.setValue(values.get(i).floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);			
					
		p.setValue(pvs);
		
		tps.setProcessingTime(p);
		ep.setTimeParameters(tps);
	}
	
	private void convertQueueTime(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
			
		Parameter p;
		if (tps.getQueueTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getQueueTime();
		}
		
		//set queue time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
				
		p.setValue(pvs);
		
		tps.setQueueTime(p);
		ep.setTimeParameters(tps);				
	}
	
	private void convertQueueTimeOrder(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		TimeParameters tps;		
		if (ep.getTimeParameters() == null) {
			tps = new TimeParameters();			
		} else {
			tps = ep.getTimeParameters();			
		}			
			
		Parameter p;
		if (tps.getQueueTime() == null) {
			p = new Parameter();
		} else {
			p = tps.getQueueTime();
		}
		
		//set queue time to p		
		Collection<ParameterValue> pvs = p.getValue();		
				
		@SuppressWarnings("unchecked")
		ArrayList<Double> values = (ArrayList<Double>) treeNode.getIndependentProperty(prop);
		float probability = 1f / values.size();
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (int i=0;i<values.size();i++){							
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
						
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);
			floatValue.setInstance(new UID().toString() + "-" + (i));
			floatValue.setValue(values.get(i).floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
					
		p.setValue(pvs);
		
		tps.setQueueTime(p);
		ep.setTimeParameters(tps);			
	}
	
	private void convertSimulationProperty(Property<?> prop, ProcessTreeElement treeNode, Scenario s) throws IllegalAccessException, InstantiationException {
		//get or create parameters	
		ScenarioParameters sp;
		if (s.getScenarioParameter() == null) {
			sp = new ScenarioParameters();
		} else {
			sp = s.getScenarioParameter();			
		}
		
		Parameter pd; //duration		
		if (sp.getDuration() == null) {			
			pd = new Parameter();
		} else {
			pd = sp.getDuration();
		}
		
		Parameter pw; //warm up period
		if (sp.getWarmUpPeriod() == null) {
			pw = new Parameter();
		} else {
			pw = sp.getWarmUpPeriod();			
		}
		
		Date start = new Date();
		start.setHours(0);
		start.setMinutes(0);
		start.setSeconds(0);
		start.setDate(1);
		start.setMonth(1);
		start.setYear(115);
		Parameter startTime;
		if(sp.getStart() == null){
			startTime = new Parameter(); 
		}
		else{
			startTime = sp.getStart();
		}
		// start time
		Collection<ParameterValue> startTimes = startTime.getValue();
		DateTimeParameter startT = new DateTimeParameter();
		startT.setValue(start);
		startTimes.add(startT);
		startTime.setValue(startTimes);
		sp.setStart(startTime);
				
		SimulationProperty simprop = (SimulationProperty)treeNode.getIndependentProperty(prop);
		
		//base time unit
		//default in Petra is hours for now
		sp.setBaseTimeUnit(TimeUnit.hour);
				
		//replication
		sp.setReplication(simprop.getNrSubRuns());
		
		//duration					
		Collection<ParameterValue> pvsd = pd.getValue();
		
		DurationParameter pvd = new DurationParameter();						
		int duration = simprop.getLengthOfSubRuns();
		pvd.setValue(new Duration.Builder().hours(duration).build());	
		pvsd.add(pvd);
		pd.setValue(pvsd);
		
		sp.setDuration(pd);
		
		//extended bpsim							
		//warm up period
		Collection<ParameterValue> pvsw = pw.getValue();
		
		DurationParameter pvw = new DurationParameter();
		int warmup = simprop.getWarmUpPeriod();
		pvw.setValue(new Duration.Builder().hours(warmup).build());			
		pvsw.add(pvw);
		pw.setValue(pvsw);
		
		sp.setWarmUpPeriod(pw);
		
		int extraSubRuns = 0;
		if (warmup > 0 && duration > 0) {
			extraSubRuns = Math.max(1, Math.round(warmup / duration));
		}
		sp.setReplication(sp.getReplication()+extraSubRuns);
		
		//push
		sp.setPush(simprop.isPush());
		
		//qType
		sp.setqType(simprop.getqType());		
				
		s.setScenarioParameter(sp);
	}
	
	private void convertWorkSchedule(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep, Scenario s) throws IllegalAccessException, InstantiationException {				
		//get or create parameters	
		// TODO: L-Sim hack
		// we need to store the amount of resources available
		Integer nrOfRes = 0;
		if(simTool.equals(SimulationTool.LSim)){
			nrOfRes = (Integer) treeNode.getIndependentProperty(NrOfResources.class);
		}
		
		
		ResourceParameters rps;		
		if (ep.getResourceParameters() == null) {
			rps = new ResourceParameters();			
		} else {
			rps = ep.getResourceParameters();			
		}
		
		Parameter p;
		if (rps.getAvailability() == null) {
			p = new Parameter();
		} else if(simTool.equals(SimulationTool.LSim)) {
			p = rps.getQuantity();
		} else {
			p = rps.getAvailability();			
		}
				 
		Collection<ParameterValue> pvs = p.getValue();
	
		//calculate sum to find frequency for recurrence
		@SuppressWarnings("unchecked")
		ArrayList<TimeValue> pal = (ArrayList<TimeValue>)treeNode.getIndependentProperty(prop);		
		double sumTime = 0;
		for (int i=0;i<pal.size();i++){
			sumTime += pal.get(i).toHours();			
		}	
		
		//set frequency of recurrence
		Frequency recurrence = Frequency.WEEKLY;
		if(sumTime == 1){
			recurrence = Frequency.HOURLY;
		} else if (sumTime == 24) {
			recurrence = Frequency.DAILY;
		} else if (sumTime == 168) {
			recurrence = Frequency.WEEKLY;
		} else if (sumTime == 672) {
			recurrence = Frequency.MONTHLY;
		}
		 
		//get or create calendar collection
		Collection<BPSimCalendar> calendars = s.getCalendar();
		if (calendars == null) {
			calendars = new HashSet<BPSimCalendar>();
		}		
				
		final long hoursInMillis = 60L * 60L * 1000L;
		
		//create calendars and availability parameters
		//Date start = new Date();
		// start at the epoch
		Date start = new Date();
		start.setHours(0);
		start.setMinutes(0);
		start.setSeconds(0);
		start.setDate(1);
		start.setMonth(1);
		start.setYear(115);
		boolean available = true;
		// first set the amount of resources to 0
		if(simTool.equals(SimulationTool.LSim)){
			NumericParameter np = new NumericParameter();
			// we have to differentiate between available and not available
			np.setValue(0l);
			pvs.add(np);
		}
		for (int i=0;i<pal.size();i++){			
			if (pal.get(i).toHours().intValue() != 0) {
				ICalendar ical = new ICalendar();
				BPSimCalendar calendar = new BPSimCalendar();
				String id = UUID.randomUUID().toString();
				calendar.setId(id);	
				
				VEvent event = new VEvent();
								
				event.setDateStart(new DateStart(start, true));
				
				long hours = pal.get(i).toHours().longValue();
				Date end = new Date(start.getTime() + (hours * hoursInMillis));
				event.setDateEnd(new DateEnd(end, true));
				//event.setDuration(new Duration.Builder().hours(pal.get(i).toHours().intValue()).build()); //all-day event
				
				//start = new Date(start.getTime() + (hours * hoursInMillis));
				start = end;						
				Recurrence recur = new Recurrence.Builder(recurrence).build();
				event.setRecurrenceRule(recur);
		
				ical.addEvent(event);
				
				calendar.setCalendar(ical);
				calendars.add(calendar);
				
				if(simTool.equals(SimulationTool.LSim) && i % 2 == 0){
					NumericParameter np = new NumericParameter();
					// we have to differentiate between available and not available
					np.setValue(nrOfRes.longValue() * (1 - (i % 2)));
					np.setValidFor(calendar);
					pvs.add(np);
				}
				else if(simTool.equals(SimulationTool.LSim)){
					
				}
				else{
					BooleanParameter pv = new BooleanParameter();
					pv.setValue(available);
					pv.setValidFor(calendar);
					pvs.add(pv);
				}
														
			}
			available = !available;						
		}//end for loop

		p.setValue(pvs);
		if(simTool.equals(SimulationTool.LSim)){
			rps.setQuantity(p);
		}
		else{
			rps.setAvailability(p);
		}
		
		TimeParameters timePar;
		if(ep.getTimeParameters() != null){
			timePar = ep.getTimeParameters();
		}
		else{
			timePar = new TimeParameters();
		}
		if(timePar.getProcessingTime() != null){
			p = timePar.getProcessingTime();
		}
		else{
			p = new Parameter();
		}
		// now for some results
		List<ResultType> lresType;
		if(p.getResultRequest() != null){
			lresType = p.getResultRequest();
		}
		else{
			lresType = new ArrayList<ResultType>();
		}
		lresType.add(ResultType.sum);
		p.setResultRequest(lresType);
		timePar.setProcessingTime(p);
		
		if(timePar.getWaitTime() != null){
			p = timePar.getWaitTime();
		}
		else{
			p = new Parameter();
		}
		// now for some results
		if(p.getResultRequest() != null){
			lresType = p.getResultRequest();
		}
		else{
			lresType = new ArrayList<ResultType>();
		}
		lresType.add(ResultType.sum);
		p.setResultRequest(lresType);
		timePar.setWaitTime(p);
		
		ep.setTimeParameters(timePar);
		
		ep.setResourceParameters(rps);
	}
	
	private void convertYerkesDodsonInterval(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		if(treeNode.getIndependentProperty(TaskDistribution.class) != null){
			// don't do both
			return;
		}
		//TODO
		//get or create parameters
		TimeParameters eps;		
		if (ep.getTimeParameters() == null) {
			eps = new TimeParameters();			
		} else {
			eps = ep.getTimeParameters();			
		}	
		
		Parameter p;
		if (eps.getProcessingTime() == null) {
			p = new Parameter();
		} else {
			p = eps.getProcessingTime();
		}
			
		Collection<ParameterValue> pvs = p.getValue();
		
		@SuppressWarnings("unchecked")
		HashMap<UUID, HashMap<Pair<Double, Double>, Distribution>> pp = (HashMap<UUID, HashMap<Pair<Double, Double>, Distribution>>)treeNode.getIndependentProperty(prop);
				
		UserDistribution pv = new UserDistribution();
		pv.setDiscrete(true);
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (UUID resource: pp.keySet()) {
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();	
			pvpoint.setProbability(1f/pp.keySet().size());
			
			if(simTool.equals(SimulationTool.LSim)){
				// now aggregate the means into a single value
				Map<Pair<Double, Double>, Distribution> cells = pp.get(resource);
				float value = 0;
				double totalInterval = 0;
				for (Pair<Double, Double> key: cells.keySet()) {
					totalInterval += key.getSecond() - key.getFirst();
				}				
				
				// so now we have the total interval, now compute the expected value
				for(Pair<Double, Double> key: cells.keySet()){
					// each FloatValue is in hours since we have L-Sim, so we can just take the float value
					value += ((key.getSecond() - key.getFirst()) / totalInterval) * getExpectedValue(cells.get(key)).getValue();
				}
				// now we have the expected value, store this in a float value
				FloatingParameter fp = new FloatingParameter();
				fp.setTimeUnit(TimeUnit.hour);
				fp.setValue(value);
				pvpoint.setValue(fp);
			}
			else{
				UserDistribution pv2 = new UserDistribution();	
				//set instance to resource id
				Object[] refLink = getBPMNElement(resource);
				boolean linkFound = (Boolean)refLink[1];
				if (linkFound) {
					NodeID linkedID = (NodeID)refLink[0];
					pv2.setInstance(linkedID.toString().replaceAll(" ", "_").toLowerCase());
				} else {				
					pv2.setInstance(resource.toString());
				}
				
				
				Map<Pair<Double, Double>, Distribution> cells = pp.get(resource);
				double totalInterval = 0;
				for (Pair<Double, Double> key: cells.keySet()) {
					totalInterval += key.getSecond() - key.getFirst();
				}
				
				Collection<UserDistributionDataPoint> points2 = pv2.getPoints();
				for (Pair<Double, Double> key: cells.keySet()) {
					UserDistributionDataPoint pvpoint2 = new UserDistributionDataPoint();	
					pvpoint2.setProbability((float)((key.getSecond() - key.getFirst())/totalInterval));				
					pvpoint2.setValue(convertDistribution(cells.get(key)));
					points2.add(pvpoint2);
				}
				pv2.setPoints(points2);
				pvpoint.setValue(pv2);
			}
			if(simTool.equals(SimulationTool.LSim)){
				// so now we have to check the pv point already has this value
				boolean found = false;
				for(UserDistributionDataPoint udd: pv.getPoints()){
					if(((FloatingParameter)(udd.getValue())).getValue().equals(((FloatingParameter)pvpoint.getValue()).getValue())){
						// update the probability of this udd
						udd.setProbability(udd.getProbability() + pvpoint.getProbability());
						found = true;
					}
				}
				if(!found){
					points.add(pvpoint);
				}
			}
			else{
				points.add(pvpoint);
			}
		}

		pvs.add(pv);
					
		p.setValue(pvs);
		
		List<ResultType> lresType = new ArrayList<ResultType>();
		lresType.add(ResultType.mean);
		p.setResultRequest(lresType);
		
		eps.setProcessingTime(p);
		
		// now also collect some extra information about the processing time
		if(simTool.equals(SimulationTool.LSim)){
			// queue time is not supported
		}
		else{
			if (eps.getQueueTime() == null) {
				p = new Parameter();
			} else {
				p = eps.getQueueTime();
			}
			lresType = new ArrayList<ResultType>();
			lresType.add(ResultType.mean);
			p.setResultRequest(lresType);
			eps.setQueueTime(p);
		}
		
		ep.setTimeParameters(eps);	
	}
	
	//-----------------------------------------------------
	//Extended BPSim
	//-----------------------------------------------------
	private void convertDataValueMatrix(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {			
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
			
		Collection<DataValueMatrixValue> p;
		if (eps.getYerkesDodsonInterval() == null) {
			p = new HashSet<DataValueMatrixValue>();
		} else {
			p = eps.getDataValueMatrix();
		}				
		
		@SuppressWarnings("unchecked")
		HashMap<Pair<String, String>, Double> pp = (HashMap<Pair<String, String>, Double>)treeNode.getIndependentProperty(prop);
				
		for (Pair<String, String> value: pp.keySet()) {
			DataValueMatrixValue dvm = new DataValueMatrixValue();
			dvm.setFirst(value.getFirst());
			dvm.setSecond(value.getSecond());
			dvm.setValue(pp.get(value));
			
			p.add(dvm);			
		}

		eps.setDataValueMatrix(p);
		ep.setExtendedParameters(eps);
	}

	private void convertDataValues(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {	
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
	
		Parameter p;
		if (eps.getDataValues() == null) {
			p = new Parameter();
		} else {
			p = eps.getDataValues();
		}
		
		Collection<ParameterValue> pvs = p.getValue();
		
		@SuppressWarnings("unchecked")
		ArrayList<String> values = (ArrayList<String>)treeNode.getIndependentProperty(prop);
		for (int i = 0; i<values.size();i++ ) {
			StringParameter pv = new StringParameter();
			pv.setValue(values.get(i));
			pvs.add(pv);
		}
		
		p.setValue(pvs);
		
		eps.setDataValues(p);

		ep.setExtendedParameters(eps);
	}
	
	private void convertFitness(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws NumberFormatException, IllegalAccessException, InstantiationException {	
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
		
		Parameter p;
		if (eps.getFitness() == null) {
			p = new Parameter();
		} else {
			p = eps.getFitness();
		}
		
		Collection<ParameterValue> pvs = p.getValue();
		
		FloatingParameter pv = new FloatingParameter();
				
		pv.setValue(Float.valueOf(treeNode.getIndependentProperty(prop).toString()));	
		pvs.add(pv);
		p.setValue(pvs);
		
		eps.setFitness(p);
		ep.setExtendedParameters(eps);
	}	
	
	private void convertORSemantics(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {			
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
	
		eps.setOrSemantic((OROptions)treeNode.getIndependentProperty(prop));

		ep.setExtendedParameters(eps);
	}
	
	private void convertThroughputTime(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
			
		Parameter p;
		if (eps.getThroughputTime() == null) {
			p = new Parameter();
		} else {
			p = eps.getThroughputTime();
		}
		
		//set throughput time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
				
		p.setValue(pvs);
		
		eps.setThroughputTime(p);
		ep.setExtendedParameters(eps);
	}
	
	private void convertThroughputTimeOrder(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
		
		Parameter p;
		if (eps.getThroughputTimeOrder() == null) {
			p = new Parameter();
		} else {
			p = eps.getThroughputTimeOrder();
		}
		
		//set throughput time to p		
		Collection<ParameterValue> pvs = p.getValue();		
				
		@SuppressWarnings("unchecked")
		ArrayList<Double> values = (ArrayList<Double>) treeNode.getIndependentProperty(prop);
		float probability = 1f / values.size();
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (int i=0;i<values.size();i++){							
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
						
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);
			floatValue.setInstance(new UID().toString() + "-" + (i));
			floatValue.setValue(values.get(i).floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
					
		p.setValue(pvs);
		
		eps.setThroughputTimeOrder(p);
		ep.setExtendedParameters(eps);	
	}
	
	private void convertUtilisation(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {			
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
			
		Parameter p;
		if (eps.getUtilisation() == null) {
			p = new Parameter();
		} else {
			p = eps.getUtilisation();
		}
		
		//set utilisation time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);

		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
				
		p.setValue(pvs);
		
		eps.setUtilisation(p);
		ep.setExtendedParameters(eps);
	}

	private void convertWaitTime(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {		
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}	
		
		Parameter p;
		if (eps.getWaitTime() == null) {
			p = new Parameter();
		} else {
			p = eps.getWaitTime();
		}
			
		//set wait time to p
		Collection<ParameterValue> pvs = p.getValue();			
		
		@SuppressWarnings("unchecked")
		HashMap<Double, Double> pp = (HashMap<Double, Double>)treeNode.getIndependentProperty(prop);

		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (Double value: pp.keySet()) {						
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
			
			float probability = pp.get(value).floatValue();
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);	
			floatValue.setValue(value.floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
				
		p.setValue(pvs);
		
		eps.setWaitTime(p);
		ep.setExtendedParameters(eps);	
	}
	
	private void convertWaitTimeOrder(Property<?> prop, ProcessTreeElement treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {
		//get or create parameters
		ExtendedParameters eps;		
		if (ep.getExtendedParameters() == null) {
			eps = new ExtendedParameters();			
		} else {
			eps = ep.getExtendedParameters();			
		}			
			
		Parameter p;
		if (eps.getWaitTimeOrder() == null) {
			p = new Parameter();
		} else {
			p = eps.getWaitTimeOrder();
		}
		
		//set wait time to p		
		Collection<ParameterValue> pvs = p.getValue();		
		
		@SuppressWarnings("unchecked")
		ArrayList<Double> values = (ArrayList<Double>) treeNode.getIndependentProperty(prop);
		float probability = 1f / values.size();
		
		UserDistribution pv = new UserDistribution();											
		
		Collection<UserDistributionDataPoint> points = pv.getPoints();
		for (int i=0;i<values.size();i++){							
			UserDistributionDataPoint pvpoint = new UserDistributionDataPoint();		
						
			pvpoint.setProbability(probability);
						
			FloatingParameter floatValue = new FloatingParameter();
			floatValue.setResult(ResultType.mean);
			floatValue.setInstance(new UID().toString() + "-" + (i));
			floatValue.setValue(values.get(i).floatValue());						
			pvpoint.setValue(floatValue);
			points.add(pvpoint);
		}
		pvs.add(pv);
					
		p.setValue(pvs);
		
		eps.setWaitTimeOrder(p);
		ep.setExtendedParameters(eps);
	}
	
//	private void convertYerkesDodsonInterval(Property<?> prop, Node treeNode, ElementParameters ep) throws IllegalAccessException, InstantiationException {			
//		//get or create parameters
//		ExtendedParameters eps;		
//		if (ep.getExtendedParameters() == null) {
//			eps = new ExtendedParameters();			
//		} else {
//			eps = ep.getExtendedParameters();			
//		}	
//		
//		Collection<YerkesDodsonInterval> p;
//		if (eps.getYerkesDodsonInterval() == null) {
//			p = new HashSet<YerkesDodsonInterval>();
//		} else {
//			p = eps.getYerkesDodsonInterval();
//		}
//								
//		@SuppressWarnings("unchecked")
//		HashMap<UUID, HashMap<Pair<Double, Double>, Distribution>> pp = (HashMap<UUID, HashMap<Pair<Double, Double>, Distribution>>)treeNode.getIndependentProperty(prop);
//				
//		for (UUID value: pp.keySet()) {
//			YerkesDodsonInterval ydi = new YerkesDodsonInterval();	
//			HashMap<Pair<Double, Double>, Distribution> cells = pp.get(value);
//			ydi.setId(value);
//			
//			Collection<YerkesDodsonIntervalValue> ydics = new HashSet<YerkesDodsonIntervalValue>();			
//			for (Pair<Double, Double> key: cells.keySet()) {
//				YerkesDodsonIntervalValue ydic = new YerkesDodsonIntervalValue();
//				ydic.setFirst(key.getFirst());
//				ydic.setSecond(key.getSecond());
//				
//				Distribution distr = cells.get(key);				
//				
//				Parameter distribution = new Parameter();
//				Collection<ParameterValue> pvs = distribution.getValue();		
//				ParameterValue pv = convertDistribution(distr);
//				pvs.add(pv);
//				distribution.setValue(pvs);
//				
//				ydic.setDistribution(distribution);
//				ydics.add(ydic);
//			}
//			
//			ydi.setValues(ydics);
//			
//			p.add(ydi);
//		}
//		
//		eps.setYerkesDodsonInterval(p);
//		ep.setExtendedParameters(eps);	
//	}
	
	//-----------------------------------------------------
	//Additional functionality
	//-----------------------------------------------------
	private ParameterValue convertDistribution(Distribution t) {						
		
		switch(t.getDistributionType()) {
			case Beta: { 
				BetaDistribution pv = new BetaDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				//float multiplier = (Integer)cTime[1]; 
				pv.setShape(t.getShape().floatValue());
				pv.setScale(t.getScale().floatValue());				
				return pv;	
			}
			case Constant: { 
				FloatingParameter pv = new FloatingParameter();
							
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				pv.setValue(t.getMean().floatValue() * multiplier);	
				return pv;
			}
			case Erlang: { 
				ErlangDistribution pv = new ErlangDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				pv.setMean(t.getMean().floatValue() * multiplier);
				pv.setK(t.getNumber()); 			
				return pv;					
			}
			case Gamma: { 
				GammaDistribution pv = new GammaDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				//float multiplier = (Integer)cTime[1]; 
				pv.setShape(t.getShape().floatValue());
				pv.setScale(t.getScale().floatValue());			
				return pv;				
			}
			case LogNormal: { 
				LogNormalDistribution pv = new LogNormalDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 				
				pv.setMean(t.getMean().floatValue() * multiplier);
				pv.setStandardDeviation(t.getVariance().floatValue() * multiplier); 				
				return pv;	
			}
			case NegativeExponential: { 				
				NegativeExponentialDistribution pv = new NegativeExponentialDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				pv.setMean(t.getMean().floatValue() * multiplier);	
				return pv;				
			}
			case Normal: {
				if(simTool.equals(SimulationTool.LSim)){
					// we need to have a truncated normal distribution
					TruncatedNormalDistribution pv = new TruncatedNormalDistribution();
					
					Object[] cTime = convertTime(t.timeGranularity);
					TimeUnit tu = (TimeUnit)cTime[0];
					pv.setTimeUnit(tu);	
					float multiplier = (Float)cTime[1]; 				
					pv.setMean(t.getMean().floatValue() * multiplier);
					pv.setStandardDeviation(t.getVariance().floatValue() * multiplier);
					pv.setMin(0f);
					// we just set the max to 1000 time the variance from the mean
					pv.setMax(t.getMean().floatValue() * multiplier + t.getVariance().floatValue() * multiplier * 1000);
					return pv;
				}
				else{
					NormalDistribution pv = new NormalDistribution();
					
					Object[] cTime = convertTime(t.timeGranularity);
					TimeUnit tu = (TimeUnit)cTime[0];
					pv.setTimeUnit(tu);	
					float multiplier = (Float)cTime[1]; 				
					pv.setMean(t.getMean().floatValue() * multiplier);
					pv.setStandardDeviation(t.getVariance().floatValue() * multiplier); 			
					return pv;
				}
			}
			case Uniform: { 
				UniformDistribution pv = new UniformDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				pv.setMin(t.getMinimum().floatValue() * multiplier);
				pv.setMax(t.getMaximum().floatValue() * multiplier);			
				return pv;				
			}
			case Weibull: { 
				WeibullDistribution pv = new WeibullDistribution();
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				pv.setTimeUnit(tu);	
				//float multiplier = (Integer)cTime[1]; 
				pv.setShape(t.getShape().floatValue());
				pv.setScale(t.getScale().floatValue());			
				return pv;	
			}			
		}
		
		return new ParameterValue();		
	}

	private Object[] convertTime(TimeOptions t) {
		// modify this to hours in case of L-Sim
		Object[] result = new Object[2];
		if(simTool.equals(SimulationTool.LSim)){
			result[0] = TimeUnit.hour;
			TimeValue tv = new TimeValueImpl(1.0);
			tv.setTimeGranularity(t);
			result[1] = tv.toHours().floatValue();
		}
		else{
			switch(t){
				case millisecond: { 
					result[0] = TimeUnit.ms;
					result[1] = 1;
					break;
				}
				case second: { 
					result[0] = TimeUnit.s;
					result[1] = 1;
					break;
				}
				case minute: { 
					result[0] = TimeUnit.min;
					result[1] = 1;
					break;
				}
				case hour: { 
					result[0] = TimeUnit.hour;
					result[1] = 1;
					break;
				}
				case day: { 
					result[0] = TimeUnit.day;
					result[1] = 1;
					break;
				}
				case week: {
					result[0] = TimeUnit.year;
					result[1] = 52;
					break;
				}
				case month: { 				
					result[0] = TimeUnit.year;
					result[1] = 12;
					break;
				}
				case year: { 
					result[0] = TimeUnit.year;
					result[1] = 1;
					break;
				}
				default: {
					result[0] = TimeUnit.s;
					result[1] = 1;
					break;
				}
			}
		}
		return result;
	}
	
	public Boolean hasElementParams(BPSim bpsim, UUID treeId){
		return getElementParams(bpsim, treeId) != null;
	}
	
	public ElementParameters getElementParams(BPSim bpsim, UUID treeId){
		for(Scenario scenario: bpsim.getBpsimData().getScenarios()){
			for(ElementParameters elemParams: scenario.getElementParameter()){
				if(elemParams.getElementRefID() != null && elemParams.getElementRefID().equals(treeId)){
					return elemParams;
				}
			}
		}
		return null;
	}
		
	protected Object[] getBPMNElement(UUID treeElement) {				
		for(NodeID key: referenceMap.keySet()) {			
			if (referenceMap.get(key).compareTo(treeElement)==0){				
				return new Object[] {key.toString(), true};
			}
		}
		if(treeElement.equals(treeId)){
			return new Object[] {bpmnTreeId, true};
		}
		return new Object[] {new NodeID().toString(), false};
	}
	
	// L-Sim specific
	private FloatingParameter getExpectedValue(Distribution t){
		FloatingParameter ret = new FloatingParameter();
		switch(t.getDistributionType()) {
			case Beta: { 				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				float multiplier = (Float)cTime[1]; 
				ret.setTimeUnit(tu);	
				ret.setValue(multiplier * (t.getShape().floatValue() / (t.getShape().floatValue() + t.getScale().floatValue())));		
				return ret;	
			}
			case Constant: { 
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				ret.setValue(t.getMean().floatValue() * multiplier);	
				return ret;
			}
			case Erlang: { 
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1];
				ret.setValue(multiplier * t.getNumber().floatValue() / (1 / t.getMean().floatValue()));
				return ret;					
			}
			case Gamma: { 
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1];
				ret.setValue(multiplier * t.getShape().floatValue() * t.getScale().floatValue());		
				return ret;				
			}
			case LogNormal: { 				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 				
				ret.setValue(multiplier * (float) Math.pow(Math.E, t.getMean() + (t.getVariance() / 2))); 				
				return ret;	
			}
			case NegativeExponential: { 					
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				ret.setValue(multiplier * t.getMean().floatValue());	
				return ret;				
			}
			case Normal: {
				// we need to have a truncated normal distribution
				
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 				
				ret.setValue(multiplier * t.getMean().floatValue());
				return ret;
			}
			case Uniform: { 	
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				ret.setValue(multiplier * (float)((t.getMaximum() + t.getMinimum()) / 2));			
				return ret;				
			}
			case Weibull: { 
				Object[] cTime = convertTime(t.timeGranularity);
				TimeUnit tu = (TimeUnit)cTime[0];
				ret.setTimeUnit(tu);	
				float multiplier = (Float)cTime[1]; 
				
				throw new NullPointerException("Weibull has not yet been implemented\n");
			}			
		}
		
		return ret;
	}
}
