package org.processmining.plugins.bpsim.exporter;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import org.apache.commons.io.IOUtils;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.bpmn.BpmnDefinitions;
import org.processmining.plugins.bpmn.BpmnDefinitions.BpmnDefinitionsBuilder;
import org.processmining.plugins.bpsim.scenario.BPSim;

public class BPSimExport {
			
private static final String ENCODING = "UTF-8";
	
	private static final String EXPORTER = "ProM. http://www.promtools.org/prom6";
	
	private static final String EXPORTER_VERSION = "6.3";
	
	private static final String TARGET_NAMESPACE = "http://www.omg.org/bpmn20";
	
	private static final String BPMN_MODEL_NAMESPACE = 
			"http://www.omg.org/spec/BPMN/20100524/MODEL";
	
	private static final String SCHEMA_LOCATION = 
			"http://www.omg.org/spec/BPMN/20100524/MODEL "
			+ "BPMN20.xsd";
	
	private static final String DI_PREFIX = "bpmndi";
	
	private static final String DI_NAMESPACE = "http://www.omg.org/spec/BPMN/20100524/DI";
	
	private static final String DC_PREFIX = "dc";
	
	private static final String DC_NAMESPACE = "http://www.omg.org/spec/DD/20100524/DC";
	
	private static final String DD_DI_PREFIX = "di";
	
	private static final String DD_DI_NAMESPACE = "http://www.omg.org/spec/DD/20100524/DI";
	
	private static final String XSI_PREFIX = "xsi";
	
	private static final String XSI_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance";
	
	private static final String CONTENT_MARKER = "@CONTENT@";
	
	private static final String BPSim_PREFIX = "bpsim";
	
	private static final String BPSim_NAMESPACE = "http://www.bpsim.org/schemas/1.0";			
	
	//first boolean: extension (true = with extension), second boolean: references (true = all references)
	public void exportPlugin(PluginContext context, BPSim bpsim, File file, boolean[] parameters) throws IOException {		
		XMLOutputFactory xmlOutput = XMLOutputFactory.newInstance();
		OutputStream outputStream = null, fileOutputStream = null;
		try {
			outputStream = new ByteArrayOutputStream();
			XMLStreamWriter writer = xmlOutput.createXMLStreamWriter(outputStream, ENCODING);
			writer.writeStartDocument(ENCODING, "1.0");			
			
			//create header
			exportEnvelope(context, writer);
			
			writer.writeEndDocument();
			writer.flush();																							
			
			String bodyString;
			//get export string of the bpmn diagram
			if(bpsim.getBpmnDefinitions() == null){
				BpmnDefinitionsBuilder definitionsBuilder = new BpmnDefinitionsBuilder(context, bpsim.getBpmnDiagram());
				BpmnDefinitions definitions = new BpmnDefinitions("definitions", definitionsBuilder);
				bodyString = definitions.exportElements();
			}
			else{
				// we have the definitions already, so lets just use that
				bodyString = bpsim.getBpmnDefinitions().exportElements();
			}
			
			//get bpsim export string
			bodyString += bpsim.exportBPSim(0, parameters);

			// Add definition content
			String envelopeString = outputStream.toString();
			String contentString = envelopeString.replace(CONTENT_MARKER, bodyString);
						
			// Write to file
			fileOutputStream = new FileOutputStream(file);
			fileOutputStream.write(contentString.getBytes());

		} catch (XMLStreamException e) {
			System.out.println("Error during export BPSim");
			e.printStackTrace();
		} finally {
			IOUtils.closeQuietly(fileOutputStream);
		}
	}	
	
	/**
	 * Export envelope
	 * 
	 * @param context
	 * @param writer
	 * @throws XMLStreamException
	 */
	private void exportEnvelope(PluginContext context, XMLStreamWriter writer)
			throws XMLStreamException {
		
		writer.writeStartElement("definitions");

		writer.setDefaultNamespace(BPMN_MODEL_NAMESPACE);
		writer.writeDefaultNamespace(BPMN_MODEL_NAMESPACE);
		
		writer.setPrefix(BPSim_PREFIX, BPSim_NAMESPACE);
		writer.writeNamespace(BPSim_PREFIX, BPSim_NAMESPACE);
		
		writer.setPrefix(DC_PREFIX, DC_NAMESPACE);
		writer.writeNamespace(DC_PREFIX, DC_NAMESPACE);
		
		writer.setPrefix(DI_PREFIX, DI_NAMESPACE);
		writer.writeNamespace(DI_PREFIX, DI_NAMESPACE);
		
		writer.setPrefix(DD_DI_PREFIX, DD_DI_NAMESPACE);
		writer.writeNamespace(DD_DI_PREFIX, DD_DI_NAMESPACE);

		writer.setPrefix(XSI_PREFIX, XSI_NAMESPACE);
		writer.writeNamespace(XSI_PREFIX, XSI_NAMESPACE);

		writer.writeAttribute("targetNamespace", TARGET_NAMESPACE);
		writer.writeAttribute("exporter", EXPORTER);
		
		writer.writeAttribute("exporterVersion", EXPORTER_VERSION);
		writer.writeAttribute(XSI_NAMESPACE, "schemaLocation", SCHEMA_LOCATION);
		
		// write marker for definition content
		writer.writeCharacters(CONTENT_MARKER);
		
		writer.writeEndElement();
	}
}
