package org.processmining.plugins.bpsim.importer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

import org.processmining.plugins.bpsim.constantparameter.FloatingParameter;
import org.processmining.plugins.bpsim.constantparameter.NumericParameter;
import org.processmining.plugins.bpsim.elementparameter.ControlParameters;
import org.processmining.plugins.bpsim.elementparameter.TimeParameters;
import org.processmining.plugins.bpsim.parameter.Parameter;
import org.processmining.plugins.bpsim.parameter.ParameterValue;
import org.processmining.plugins.bpsim.parameter.ResultType;
import org.processmining.plugins.bpsim.scenario.BPSim;
import org.processmining.plugins.bpsim.scenario.ElementParameters;
import org.processmining.plugins.bpsim.scenario.Scenario;
import org.processmining.plugins.bpsim.scenario.VendorExtension;


/**
 * This will import a BPSim result file (need to have based for the implemented elements)
 * @author DSchunse
 *
 */
public class BPSimImport {
	
	public BPSim importResults(BPSim bpsim, File file) throws IOException{
		String BPSimContents = readInFile(file);
		parseBPSimResult(bpsim, BPSimContents);
		return bpsim;
	}
	
	private void parseBPSimResult(BPSim bpsim, String result){
		// let's first get the scenario
		Scenario scenario = parseScenario(bpsim, result);
		// we have the entire result thing
		String[] element = result.split("<bpsim:ElementParameters");
		for(int i = 1; i < element.length; i++){
			parseElementParameters(scenario, element[i]);
		}
	}
	
	private Scenario parseScenario(BPSim bpsim, String result){
		String scenarioId = result.split("result=\"")[1].split("\"")[0];
		scenarioId = scenarioId.replace("_", " ");
		for(Scenario scenario: bpsim.getBpsimData().getScenarios()){
			if(scenario.getId().equalsIgnoreCase(scenarioId)){
				return scenario;
			}
		}
		assert false;
		return null;
	}
	
	private void parseElementParameters(Scenario scenario, String result){
		// we have the element here, let's first get the id of the node
		String nodeId = result.split("elementRef=\"")[1].split("\">")[0].replace("_", " ");
		if(nodeId.contains("proc")){
			nodeId = nodeId.replace(" ", "_");
		}
		// now lets get the element parameters
		ElementParameters elemParams = null;
		for(ElementParameters elemParamsIt: scenario.getElementParameter()){
			if(elemParamsIt.getElementRefnodeID() != null && elemParamsIt.getElementRefnodeID().equalsIgnoreCase(nodeId)){
				elemParams = elemParamsIt;
			}
		}
		// now let's do the different result types
		String timeParameters = "";
		if(result.contains("<bpsim:TimeParameters>")){
			timeParameters = result.split("<bpsim:TimeParameters>")[1].split("</bpsim:TimeParameters>")[0];
		}
		String controlParameters = "";
		if(result.contains("<bpsim:ControlParameters>")){
			controlParameters = result.split("<bpsim:ControlParameters>")[1].split("</bpsim:ControlParameters>")[0];
		}
		
		String vendorExtensions = "";
		if(result.contains("<bpsim:VendorExtension")){
			vendorExtensions = result.split("<bpsim:VendorExtension")[1].split("</bpsim:VendorExtension>")[0];
		}
		
		if(timeParameters != "" && elemParams != null){
			TimeParameters timeParams = parseTimeParameters(timeParameters);
			elemParams.setTimeParameters(timeParams);
		}
		if(controlParameters != "" && elemParams != null){
			ControlParameters contrParams = parseControlParameters(controlParameters);
			elemParams.setControlParameters(contrParams);
		}
		if(vendorExtensions != "" && elemParams != null){
			Collection<VendorExtension> venExtensions = parseVendorExtensions(vendorExtensions);
			elemParams.setVendorExtension(venExtensions);
		}
	}
	
	private TimeParameters parseTimeParameters(String result){
		TimeParameters timeParams = new TimeParameters();
		if(result.contains("<bpsim:ProcessingTime>")){
			String processingTime = result.split("<bpsim:ProcessingTime>")[1].split("</bpsim:ProcessingTime>")[0]; 
			timeParams.setProcessingTime(parseTime(processingTime));
		}
		if(result.contains("<bpsim:WaitTime>")){
			String waitTime = result.split("<bpsim:WaitTime>")[1].split("</bpsim:WaitTime>")[0]; 
			timeParams.setWaitTime(parseTime(waitTime));
		}
		return timeParams;
	}
	
	private ControlParameters parseControlParameters(String result){
		ControlParameters contrParams = new ControlParameters();
		String triggerCount = result.split("<bpsim:TriggerCount>")[1].split("</bpsim:TriggerCount>")[0]; 
		contrParams.setTriggerCount(parseTriggerCount(triggerCount));
		return contrParams;
	}
	
	//TODO: make this more robust
	private Collection<VendorExtension> parseVendorExtensions(String result){
		// we start with the name
		String name = result.split("name=\"")[1].split("\"")[0];
		// the rest can just go as the object
		String contents = result.split("name=\"")[1].split("\">")[1];
		VendorExtension venEx = new VendorExtension();
		venEx.setName(name);
		venEx.setValue(contents);
		Collection<VendorExtension> venExtensions = new ArrayList<>();
		venExtensions.add(venEx);
		return venExtensions;
		
	}
	
	private Parameter parseTime(String result){
		Parameter processingTime = new Parameter();
		Collection<ParameterValue> values = new ArrayList<ParameterValue>();
		String[] floatingPointParameters = result.split("<bpsim:FloatingParameter");
		for(int i = 1; i < floatingPointParameters.length; i++){
			values.add(parseFloatingPointParameter(floatingPointParameters[i]));
		}
		processingTime.setValue(values);
		return processingTime;
	}
	
	private Parameter parseTriggerCount(String result){
		Parameter triggerCount = new Parameter();
		Collection<ParameterValue> values = new ArrayList<ParameterValue>();
		String[] numericParameters = result.split("<bpsim:NumericParameter");
		for(int i = 1; i < numericParameters.length; i++){
			values.add(parseNumericParameter(numericParameters[i]));
		}
		triggerCount.setValue(values);
		return triggerCount;
	}
	
	private FloatingParameter parseFloatingPointParameter(String result){
		// get the value, type, and instance
		String value = result.split("value=\"")[1].split("\"")[0];
		String resultType = result.split("result=\"")[1].split("\"")[0];
		String instance = result.split("instance=\"")[1].split("\"")[0];
		
		FloatingParameter param = new FloatingParameter();
		param.setValue(Float.valueOf(value));
		param.setResult(ResultType.valueOf(resultType));
		param.setInstance(instance);
		return param;
	}
	
	private NumericParameter parseNumericParameter(String result){
		// get the value, type, and instance
		String value = result.split("value=\"")[1].split("\"")[0];
		String resultType = result.split("result=\"")[1].split("\"")[0];
		String instance = result.split("instance=\"")[1].split("\"")[0];
		
		NumericParameter param = new NumericParameter();
		// we have to deviate from here since L-SIM does not work properly here
		//param.setValue(Long.valueOf(value));
		param.setValue(Double.valueOf(value).longValue());
		param.setResult(ResultType.valueOf(resultType));
		param.setInstance(instance);
		return param;
	}
	
	private String readInFile(File file) throws IOException{
		BufferedReader read = new BufferedReader(new FileReader(file));
		String ret = "";
		String line;
		while((line = read.readLine()) != null){
			ret += line;
		}
		return ret;
	}

}
